unit stexp;

interface
uses dialogs,classes,SysUtils;

type TAddProperty = (proWrite,ProRead,proFRead,proFWrite);
type TAddPropertySet = set of TAddProperty;

function upos(const sub,st:string):integer;
function FindClassLine(const Lines:TStrings; start:integer): integer;
function FindLastPosFor(const Lines:TStrings; const cn:string): integer;
function FindLine(const Lines:TStrings; const cn:string): integer;
function IsMethod(const s:string):integer;
function GetClassName(const s:string):string;
function MakeFullMethod(const s,cn:string):string;
function MakeMethod(const s,proc:string):string;
function GetMethod(const Lines:TStrings; ln:integer) : string;
function GetNewClassMethod(const Lines:TStrings; ln:integer;
                           const proc : string; var InsertPos:integer) : string;
function AddPropertyUnder(const Lines:TStrings; ln:integer; typs : TAddPropertySet) : boolean;


implementation

const crlf  = #13#10;
const crlf2 = crlf+crlf;
const commentline =
'//-----------------------------------------------------------------------------'
+crlf;

//------------------------------------------------------------------------------
function upos(const sub,st:string):integer;
begin
  Result := pos(sub,st);
end;


function FindClassLine(const Lines:TStrings; start:integer): integer;
var i,p1,p2:integer;
begin
  i := start;
  Result := -1;
  if ( i < 0 ) then exit;
  while ( true ) do begin
    p1 := upos('class',Lines[i]);   // Must be TName = class(...
    if ( p1 <> 0 ) then begin
      p2 := pos('=',Lines[i]);
      if ( p2 > 0 ) and ( p2 < p1 ) then break;
    end;
    dec(i);
    if ( i < 0 ) then exit;
    if ( upos('end;',Lines[i]) <> 0 ) then exit;
  end;
  Result := i;
end;

function FindLastPosFor(const Lines:TStrings; const cn:string): integer;
var i:integer;
begin
  i := Lines.Count-1;
  Result := -1;
  if ( i < 0 ) then exit;
  while ( upos('end;',Lines[i]) <> 1 ) do begin
    dec(i);
    if ( i < 0 ) then exit;
    Result := i;
    if ( upos('implementation',Lines[i]) <> 0 ) then exit;
    if ( upos('{$R *.DFM}',Lines[i]) <> 0 ) then exit;
  end;
  Result := i;
end;

function FindLine(const Lines:TStrings; const cn:string): integer;
var i:integer;  pn : string;
begin
  pn := cn; i := pos('//',pn); if ( i > 0 ) then Delete(pn,i,1000);
  pn := trim(pn);
  i := Lines.Count-1;
  Result := -1;
  if ( i < 0 ) then exit;
  while ( pos(pn,Lines[i]) = 0 ) do begin
    dec(i);
    if ( i < 0 ) then exit;
  end;
  Result := i;
end;

function IsMethod(const s:string):integer;
begin
  if ( upos('procedure',s) > 0 )   then begin Result := 1; exit; end;
  if ( upos('function',s) > 0 )    then begin Result := 2; exit; end;
  if ( upos('constructor',s) > 0 ) then begin Result := 3; exit; end;
  if ( upos('destructor',s) > 0 )  then begin Result := 4; exit; end;
  Result := 0;
end;

function GetClassName(const s:string):string;
var p:integer;
begin
  Result := trim(s);
  p := pos('=',Result);
  if ( p < 1 ) then exit;
  Delete(Result,p,200);
  Result := trim(Result);
end;

function MakeFullMethod(const s,cn:string):string;
var p,p2:integer;
begin
  Result := trim(s);
  p := pos(' ',Result);
  if ( p = 0 ) then exit;
  while ( Result[p] = ' ' ) do inc(p);
  Insert(cn+'.',Result,p);
  p  := pos(')',Result);
  p2 := pos(';',Result);
  if ( p > p2 ) then p2 := p;
  p := pos(';',copy(Result,p2,1000));
  if ( p+p2 < Length(Result) ) then insert(' // ',Result,p2+p);
end;

function MakeMethod(const s,proc:string):string;
var m:integer;
begin
  m := IsMethod(s);
  Result := '';
  case m of
    1 : Result := crlf2 + commentline +
                  s + crlf +
                  'begin' + crlf +
                  '  ' + proc + crlf +
                  'end;' + crlf2;
    2 : Result := crlf2  + commentline +
                  s + crlf +
                  'begin' + crlf +
                  '  Result := ' + proc + crlf +
                  'end;' + crlf2;
    3 : Result := crlf2 + commentline +
                  s + crlf +
                  'begin' + crlf +
                  '  inherited;' + crlf +
                  '  ' + proc + crlf +
                  'end;' + crlf2;
    4 : Result := crlf2 + commentline +
                  s + crlf +
                  'begin' + crlf +
                  '  ' + proc + crlf +
                  '  inherited;' + crlf +
                  'end;' + crlf2;
  end;
end;

function GetMethod(const Lines:TStrings; ln:integer) : string;
begin
  Result := Lines[ln];
end;

function FindClassTextLine(const Lines:TStrings; ln:integer;const s:string):integer;
var i:integer;
begin
  Result := -1;
  for i:=ln to Lines.Count-1 do begin
    if ( upos(s,Lines[i]) > 0 ) then begin Result := i; exit; end;
    if ( upos('end;',Lines[i]) > 0 ) then begin Result := i; exit; end;
  end;
end;

function GetNewClassMethod(const Lines:TStrings; ln:integer;
                           const proc : string; var InsertPos:integer) : string;
var  oldline,ClassPosLine : integer;
     CurrentClassName,CurrentLine,MethodProtoType : string;

begin
  Result := '';
  CurrentLine := GetMethod(Lines,ln);
  if ( IsMethod(CurrentLine) < 1 ) then
    raise Exception.Create('Must be over method line');
  ClassPosLine := FindClassLine(Lines,ln);
  if ( ClassPosLine < 0 ) then
    raise Exception.Create('Must be inside a class');
  CurrentClassName := GetClassName(Lines[ClassPosLine]);
  MethodProtoType := MakeFullMethod(CurrentLine,CurrentClassName);
  InsertPos := FindLastPosFor(Lines,CurrentClassName);
  oldline := FindLine(Lines,MethodProtoType);
//    ShowMessage(IntToStr(oldline)+' '+MakeMethod(MethodProtoType));
  if ( oldline >= 0 ) then exit;
  Result := MakeMethod(MethodProtoType,proc);
end;

function AddMethodUnder(const Lines:TStrings; ln:integer;
                        const proc : string) : boolean;
var InsertPos : integer; NewMethod : string;
    NewLines : TStrings;
    i : integer;
begin
  Result := false;
  NewMethod := GetNewClassMethod(Lines,ln,proc,InsertPos);
  if ( NewMethod = '' ) then exit;
  NewLines := TStringList.Create;
  NewLines.Text := NewMethod;
  for i:=0 to NewLines.Count-1 do
    Lines.Insert(InsertPos+1+i,NewLines[i]);
  NewLines.Free;
  Result := true;
end;

function AddPropertyUnder(const Lines:TStrings; ln:integer; typs : TAddPropertySet) : boolean;
var  ClassPosLine, ClassEndLine : integer;
     CurrentClassName,CurrentLine : string;
     prop,proptype,spaces : string;
     propline : string;
     fp,sp : integer;
begin
  Result := false;
  CurrentLine := Lines[ln];
  ClassPosLine := FindClassLine(Lines,ln);
  if ( ClassPosLine < 0 ) then
    raise Exception.Create('Must be inside a class');
  CurrentClassName := GetClassName(Lines[ClassPosLine]);

  fp := upos('F',CurrentLine);
  if ( fp = 0 ) then
    raise Exception.Create('Must be over property attribute starting by F');
  sp := pos(':',CurrentLine);
  prop := trim(Copy(CurrentLine,fp+1,sp-fp-1));
  proptype := trim(Copy(CurrentLine,sp+1,1000));
  sp := pos(';',proptype); delete(proptype,sp,1000);
  spaces := Copy(CurrentLine,1,fp-1);

  ClassPosLine := FindClassLine(Lines,ln);
  if ( ClassPosLine < 0 ) then
    raise Exception.Create('Must be inside a class');
  CurrentClassName := GetClassName(Lines[ClassPosLine]);

  propline := spaces + 'property ' + prop + ' : ' + proptype;
  if ( proRead in typs )   then propline := propline + ' read Get' + prop;
  if ( proFRead in typs )  then propline := propline + ' read F' + prop;
  if ( proWrite in typs )  then propline := propline + ' write Set' + prop;
  if ( proFWrite in typs ) then propline := propline + ' write F' + prop;
  propline := propline + ';';

  ClassEndLine := FindClassTextLine(Lines,ClassPosLine,'end;');
  Lines.Insert(ClassEndLine,propline);
  ClassEndLine := FindClassTextLine(Lines,ClassPosLine,'published');
  if ( proWrite in typs ) then begin
    Lines.Insert(ClassEndLine  ,spaces + 'procedure Set' + prop + '(value:'
                 + proptype +'); virtual;');
    AddMethodUnder(Lines,ClassEndLine  ,'F'+prop+' := value;');
  end;
  if ( proRead in typs ) then begin
    Lines.Insert(ClassEndLine,spaces + 'function  Get' + prop + ' : '
               + proptype +'; virtual;');
    AddMethodUnder(Lines,ClassEndLine,'F'+prop+';');
  end;
//  ShowMessage(Lines.Text);
  Result := true;
end;

end.
